#ifndef _UTIL_LOCK_H
#define _UTIL_LOCK_H

// There are two basic lock types: Lock and SpinLock. A thread that
// blocks on a Lock may sleep until it obtains the lock.  In kernel code,
// a thread that blocks on a SpinLock will not sleep, but will run a busy
// loop until the lock is freed.  In userspace code, a SpinLock behaves
// as an ordinary Lock.

#ifdef _KERNEL
#include <kern/lock.h>
#endif

#include <util/spinlock.h>

namespace Lock {
	struct AutoLock {
		Lock &real_lock;
	
		AutoLock(Lock &lock) : real_lock(lock)
		{
			real_lock.lock();
		}
		
		~AutoLock()
		{
			real_lock.unlock();
		}
	};
	
	struct DroppableAutoLock {
		Lock &real_lock;
		bool dropped;
		
		DroppableAutoLock(Lock &lock) : real_lock(lock)
		{
			real_lock.lock();
			dropped = false;
		}
		
		~DroppableAutoLock()
		{
			if (!dropped)
				real_lock.unlock();
		}
		
		void lock()
		{
			assert(dropped);
			real_lock.lock();
			dropped = false;
		}
		
		void unlock()
		{
			assert(!dropped);
			dropped = true;
			real_lock.unlock();
		}
	};

	struct DoubleAutoLock {
		Lock &real_lock1, &real_lock2;
	
		DoubleAutoLock(Lock &lock1, Lock &lock2) :
		real_lock1(lock1), real_lock2(lock2)
		{
			if (&lock1 < &lock2) {
				real_lock1.lock();
				real_lock2.lock();
			} else {
				real_lock2.lock();
				real_lock1.lock();
			}
		}
		
		~DoubleAutoLock()
		{
			real_lock1.unlock();
			real_lock2.unlock();
		}
	};
}

#endif
