#ifndef _UTIL_SPINLOCK_H
#define _UTIL_SPINLOCK_H

// There are two basic lock types: Lock and SpinLock. A thread that
// blocks on a Lock may sleep until it obtains the lock.  In kernel code,
// a thread that blocks on a SpinLock will not sleep, but will run a busy
// loop until the lock is freed.  In userspace code, a SpinLock behaves
// as an ordinary Lock.

#ifdef _KERNEL
#include <kern/spinlock.h>
#endif

namespace Lock {
#ifndef _KERNEL
#error FIXME -- implement locks for userspace
	struct Lock {
	};

	typedef Lock SpinLock;
#endif

	struct AutoSpinLock {
		SpinLock &real_lock;
	
		AutoSpinLock(SpinLock &lock) : real_lock(lock)
		{
			real_lock.lock();
		}
		
		~AutoSpinLock()
		{
			real_lock.unlock();
		}
	};

	struct DroppableAutoSpinLock {
		SpinLock &real_lock;
		bool dropped;
		
		DroppableAutoSpinLock(SpinLock &lock) : real_lock(lock)
		{
			real_lock.lock();
			dropped = false;
		}
		
		~DroppableAutoSpinLock()
		{
			if (!dropped)
				real_lock.unlock();
		}
		
		void lock()
		{
			assert(dropped);
			real_lock.lock();
			dropped = false;
		}
		
		void unlock()
		{
			assert(!dropped);
			dropped = true;
			real_lock.unlock();
		}
	};

	struct AutoSpinLockIRQ {
		SpinLock &real_lock;
	
		AutoSpinLockIRQ(SpinLock &lock) : real_lock(lock)
		{
			real_lock.lock_irq();
		}
		
		~AutoSpinLockIRQ()
		{
			real_lock.unlock_irq();
		}
	};

	struct DroppableAutoSpinLockIRQ {
		SpinLock &real_lock;
		bool dropped;
		
		DroppableAutoSpinLockIRQ(SpinLock &lock) : real_lock(lock)
		{
			real_lock.lock_irq();
			dropped = false;
		}
		
		~DroppableAutoSpinLockIRQ()
		{
			if (!dropped)
				real_lock.unlock_irq();
		}
		
		void lock()
		{
			assert(dropped);
			real_lock.lock_irq();
			dropped = false;
		}
		
		void unlock()
		{
			assert(!dropped);
			dropped = true;
			real_lock.unlock_irq();
		}
	};

	struct AutoSpinLockRecIRQ {
		SpinLock &real_lock;
		ulong savedirq;
	
		AutoSpinLockRecIRQ(SpinLock &lock) : real_lock(lock)
		{
			savedirq = real_lock.lock_recirq();
		}
		
		~AutoSpinLockRecIRQ()
		{
			real_lock.unlock_recirq(savedirq);
		}
	};

	struct DroppableAutoSpinLockRecIRQ {
		SpinLock &real_lock;
		ulong savedirq;
		bool dropped;
		
		DroppableAutoSpinLockRecIRQ(SpinLock &lock) : real_lock(lock)
		{
			savedirq = real_lock.lock_recirq();
			dropped = false;
		}
		
		~DroppableAutoSpinLockRecIRQ()
		{
			if (!dropped)
				real_lock.unlock_recirq(savedirq);
		}
		
		void lock()
		{
			assert(dropped);
			savedirq = real_lock.lock_recirq();
			dropped = false;
		}
		
		void unlock()
		{
			assert(!dropped);
			dropped = true;
			real_lock.unlock_recirq(savedirq);
		}
	};
}

#endif
