// Byte-swapping, unaligned memory, and I/O port access functions. These
// functions are all unprivileged, except the ll_in_* and ll_out_*
// functions, whose privilege is architecture-dependent (on PPC and most
// other non-x86 architectures, they require the address specified to be
// mapped appropriately).

#ifndef _LL_ARCH_IO_H
#define _LL_ARCH_IO_H

#include <lowlevel/types.h>

// Unconditional byte-swap

static inline uint16_t ll_swap16(uint16_t val)
{
	return (uint16_t)val | ((uint16_t)val >> 8) << 8);
}

static inline uint32_t ll_swap32(uint32_t val)
{
	return ((val & 0x000000ff) << 24) |
	       ((val & 0x0000ff00) <<  8) |
   	    ((val & 0x00ff0000) >>  8) |
      	 ((val & 0xff000000) >> 24);
}

static inline uint64_t ll_swap64(uint64_t val)
{
	return (uint64_t)ll_swap32(val) | ((uint64_t)ll_swap32(val >> 32) << 32);
}

// Conditional byte-swap to/from a specific endianness

static inline uint64_t ll_swap_be64(uint64_t val)
{
	return val;
}

static inline uint32_t ll_swap_be32(uint32_t val)
{
	return val;
}

static inline uint16_t ll_swap_be16(uint16_t val)
{
	return val;
}

static inline uint64_t ll_swap_le64(uint64_t val)
{
	return ll_swap64(val);
}

static inline uint32_t ll_swap_le32(uint32_t val)
{
	return ll_swap32(val);
}

static inline uint16_t ll_swap_le16(uint16_t val)
{
	return ll_swap16(val);
}

// Unaligned, host-endian

static inline uint16_t ll_load_unaligned_16(void *addr)
{
	return *(uint16_t *)addr;
}

static inline uint64_t ll_load_unaligned_64(void *addr)
{
	// FIXME: atomic using FP?
	return *(uint64_t *)addr;
}

static inline uint32_t ll_load_unaligned_32(void *addr)
{
	return *(uint32_t *)addr;
}

static inline void ll_store_unaligned_16(void *addr, uint16_t val)
{
	*(uint16_t *)addr = val;
}

static inline void ll_store_unaligned_64(void *addr, uint64_t val)
{
	// FIXME: atomic using FP?
	*(uint64_t *)addr = val;
}

static inline void ll_store_unaligned_32(void *addr, uint32_t val)
{
	*(uint32_t *)addr = val;
}

// Unaligned, little-endian

static inline uint16_t ll_load_unaligned_le16(void *addr)
{
	uint16_t val;
	asm("lhbrx %0, 0, %1" : "=r" (val) : "r" (addr), "m" (*(uint16_t *)addr));
	return val;
}

static inline uint32_t ll_load_unaligned_le32(void *addr)
{
	uint32_t val;
	asm("lwbrx %0, 0, %1" : "=r" (val) : "r" (addr), "m" (*(uint32_t *)addr));
	return val;
}

static inline uint64_t ll_load_unaligned_le64(void *addr)
{
	// FIXME: atomic using FP?
	return (uint64_t)ll_load_unaligned_le32(addr) |
	       ((uint64_t)ll_load_unaligned_le32((void *)((char *)addr + 4)) << 32);
}

static inline void ll_store_unaligned_le16(void *addr, uint16_t val)
{
	asm("sthbrx %1, 0, %2" : "=m" (*(uint16_t *)addr) : "r" (val), "r" (addr));
}

static inline void ll_store_unaligned_le32(void *addr, uint32_t val)
{
	asm("stwbrx %1, 0, %2" : "=m" (*(uint32_t *)addr) : "r" (val), "r" (addr));
}

static inline void ll_store_unaligned_le64(void *addr, uint64_t val)
{
	// FIXME: atomic using FP?
	ll_store_unaligned_le32(addr, (uint32_t)val);
	ll_store_unaligned_le32((void *)((char *)addr + 4), (uint32_t)(val >> 32));
}

// Aligned, little-endian

static inline uint64_t ll_load_le64(uint64_t *addr)
{
	return ll_load_unaligned_le64(addr);
}

static inline uint32_t ll_load_le32(uint32_t *addr)
{
	return ll_load_unaligned_le32(addr);
}

static inline uint16_t ll_load_le16(uint16_t *addr)
{
	return ll_load_unaligned_le16(addr);
}

static inline void ll_store_le64(uint64_t *addr, uint64_t val)
{
	ll_store_unaligned_le64(addr, val);
}

static inline void ll_store_le32(uint32_t *addr, uint32_t val)
{
	ll_store_unaligned_le32(addr, val);
}

static inline void ll_store_le16(uint16_t *addr, uint16_t val)
{
	ll_store_unaligned_le16(addr, val);
}

// Unaligned, big-endian

static inline uint64_t ll_load_unaligned_be64(void *addr)
{
	return ll_load_unaligned_64(addr);
}

static inline uint32_t ll_load_unaligned_be32(void *addr)
{
	return ll_load_unaligned_32(addr);
}

static inline uint16_t ll_load_unaligned_be16(void *addr)
{
	return ll_load_unaligned_16(addr);
}

static inline void ll_store_unaligned_be32(void *addr, uint32_t val)
{
	ll_store_unaligned_32(addr, val);
}

static inline void ll_store_unaligned_be16(void *addr, uint16_t val)
{
	ll_store_unaligned_16(addr, val);
}

// Aligned, big-endian

static inline uint32_t ll_load_be32(uint32_t *addr)
{
	return ll_load_unaligned_be32(addr);
}

static inline uint16_t ll_load_be16(uint16_t *addr)
{
	return ll_load_unaligned_be16(addr);
}

static inline void ll_store_be32(uint32_t *addr, uint32_t val)
{
	ll_store_unaligned_be32(addr, val);
}

static inline void ll_store_be16(uint16_t *addr, uint16_t val)
{
	ll_store_unaligned_be16(addr, val);
}

// PCI/ISA/similar I/O-space; if memory-mapped, addr must include
// the base of the I/O window.

static inline uint8_t ll_in_8(ulong addr)
{
	return *(uint8_t *)addr;
}

static inline uint16_t ll_in_be16(ulong addr)
{
	return ll_load_be16(addr);
}

static inline uint32_t ll_in_be32(ulong addr)
{
	return ll_load_be32(addr);
}

static inline uint16_t ll_in_le16(ulong addr)
{
	return ll_load_le16(addr);
}

static inline uint32_t ll_in_le32(ulong addr)
{
	return ll_load_le32(addr);
}

static inline void ll_out_8(ulong addr, uint8_t val)
{
	*(uint8_t *)addr = val;
}

static inline void ll_out_be16(ulong addr, uint16_t val)
{
	ll_store_be16(addr, val);
}

static inline void ll_out_be32(ulong addr, uint32_t val)
{
	ll_store_be32(addr, val);
}

static inline void ll_out_le16(ulong addr, uint16_t val)
{
	ll_store_le16(addr, val);
}

static inline void ll_out_le32(ulong addr, uint32_t val)
{
	ll_store_le32(addr, val);
}
	
#endif
