// arch/x86/descriptors.cc -- code to manage segment/trap/interrupt descriptors
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <arch/addrs.h>
#include <arch/thread.h>
#include <arch/mem.h>

namespace Arch {
namespace Priv {
	TSS tss;
}
}

using Arch::Priv::Descriptor;
using Arch::Priv::TSS;
using Arch::Priv::tss;

Descriptor x86_gdt[1024] = {
	{}, // The first entry is reserved for the NULL selector.
	{   // 0x08: data
		limit_low: 0xffff,
		base_low: 0,
		base_mid: 0,
		type: 2, // data segment, writable
		user: 1,
		dpl: 0,
		present: 1,
		limit_high: 0xff,
		sw: 0,
		reserved: 0,
		opsize: 1,
		gran: 1,
		base_high: 0
	},
	{   // 0x10: code
		limit_low: 0xffff,
		base_low: 0,
		base_mid: 0,
		type: 10, // code segment, readable
		user: 1,
		dpl: 0,
		present: 1,
		limit_high: 0xff,
		sw: 0,
		reserved: 0,
		opsize: 1,
		gran: 1,
		base_high: 0
	},
	{   // 0x18: TSS
		limit_low: sizeof(tss),
		base_low: 0,
		base_mid: 0,
		type: 9,
		user: 0,
		dpl: 0,
		present: 1,
		limit_high: 0,
		sw: 0,
		reserved: 0,
		opsize: 0,
		gran: 0,
		base_high: 0
	}
};

struct X86DescriptorTablePointer {
	u8 pad[6];
	u16 limit;
	u32 address;
};

X86DescriptorTablePointer x86_gdtr = {
	// G++ still won't handle complex labelled initializers, so
	// we have to explicitly initialize pad.
	
	pad: { 0, 0, 0, 0, 0, 0 }, 
	limit: sizeof(x86_gdt),
	address: reinterpret_cast<u32>(&x86_gdt)
};

X86DescriptorTablePointer x86_gdtr_phys = {
	pad: { 0, 0, 0, 0, 0, 0 }, 
	limit: sizeof(x86_gdt),
	address: reinterpret_cast<u32>(&x86_gdt) - KERNEL_START
};

struct X86InterruptDescriptor {
	u16 offset_low;
	u16 selector;
	u8 stack_index:3;
	u8 reserved:5;
	u8 type:4;
	u8 user:1;
	u8 dpl:2;
	u8 present:1;
	u16 offset_high;
};

static X86InterruptDescriptor idt[256];

static X86DescriptorTablePointer idtr = {
	pad: { 0, 0, 0, 0, 0, 0 }, 
	limit: sizeof(idt),
	address: reinterpret_cast<u32>(&idt)
};

// Set a gate for INT num to start executing at addr.  
//
// If ints_off is set, then interrupts will remain disabled until
// software enables them; otherwise, the interrupt flag will have the
// same state as in the code that was interrupted.
//
// If user is true, then this gate can be called directly from
// userspace with the INT instruction.  Otherwise, usermode
// INT to this gate will cause a GPF, but it may still be reached
// from user code via hardware interrupt or exception.
// 
// If stack_index is non-zero, use the specified alternate stack even if
// interrupting kernel code.

static void set_int_gate(int num, void *addrptr, bool ints_off = false,
                         bool user = false, int stack_index = 0)
{
	u32 addr = (u32)addrptr;

	X86InterruptDescriptor desc = {
		offset_low: addr & 0xffff,
		selector: 0x10,
		stack_index: stack_index,
		reserved: 0,
		type: ints_off ? 14 : 15,
		user: 0,
		dpl: user ? 3 : 0,
		present: 1,
		offset_high: (addr >> 16) & 0xffff,
	};
	
	idt[num] = desc;
}

extern int x86_diverr, x86_debug, x86_breakpoint;
extern int x86_gpf, x86_page_fault, x86_invalid_insn, x86_int99_entry;
extern void *x86_irqs[256];

namespace Arch {
namespace Priv {
	void set_idt()
	{
		// GCC 4.0 pukes on "m" (&idtr.limit), saying it's not
		// directly addressable.
		
		asm volatile("lidtl 6(%0)" : : "r" (&idtr) : "memory");
		
		set_int_gate(0, &x86_diverr);
		set_int_gate(1, &x86_debug);
		set_int_gate(3, &x86_breakpoint);
		set_int_gate(6, &x86_invalid_insn);
		set_int_gate(13, &x86_gpf);
		set_int_gate(14, &x86_page_fault, true);
		set_int_gate(0x99, &x86_int99_entry);
		
		for (int i = 0x20; i < 0x30; i++)
			set_int_gate(i, x86_irqs[i], true);
	}
}
}
