#ifndef _KERN_IRQ_H
#define _KERN_IRQ_H

#include <kern/kernel.h>
#include <util/spinlock.h>

#include <System/IO/Interrupts.h>

namespace IRQ {
	using namespace Lock;

	typedef System::IO::Interrupts::Interrupt IInterrupt;
	typedef System::IO::Interrupts::UserInterrupt IUserInterrupt;
	typedef System::IO::Interrupts::InterruptController IInterruptController;
	using System::IO::Bus::Device;
	
	struct Interrupt;
	class InterruptController;
	
	struct InterruptSlot {
		InterruptController *controller;
		Interrupt *first_int;
		int mask_count;
		
		static const u32 Pending = 0x00000001;
		static const u32 Running = 0x00000002;
		
		u32 flags;
	};

	class InterruptController {
		// When this function returns, any irq instances that were running
		// when the function was entered will have completed.  If not
		// masked, though, there could be new instances of the IRQ running
		// when this function exits.
		
		void wait_for_irq(InterruptSlot *slot);
	
	protected:
		SpinLock lock;
	
		// The derived constructor must initialize these.
		InterruptSlot *irqs;
		u32 num_irqs;

		virtual void mask_and_ack(u32 irq) = 0;
		virtual void mask(u32 irq) = 0;
		virtual void unmask(u32 irq) = 0;

		// Returns the highest-priority pending, unmasked interrupt, or
		// negative if no pending unmasked interrupts.
		
		virtual int get_pending_irq() = 0;
			
	public:
//		#include "core/irq-server/IRQ/InterruptController.h"
		
		virtual ~InterruptController()
		{
		}
		
		void request_int(InterruptSlot *slot, Interrupt *irq);
		void free_int(Interrupt *irq);
	
		// If irq is non-negative, then the given IRQ number is the IRQ
		// source (this is a hack for x86 and x64, which provide this
		// information via the entry point used).  Otherwise,
		// get_pending_irq() will be called repeatedly (after handling
		// whatever interrupt was returned on the previous iteration) until
		// it returns negative.  This returns true if any of the handlers
		// called returned true.
	
		bool handle_irq(int irq);
		
		// Recursively request masking/unmasking of an interrupt.

		void rec_mask(InterruptSlot *slot);
		void rec_unmask(InterruptSlot *slot);

		// Like rec_mask, but does not wait for the completion of any pending
		// or running handlers.
		
		void rec_mask_nowait(InterruptSlot *slot);
		
		InterruptSlot *get_slot(u32 irq)
		{
			assert(irq < num_irqs);
			return &irqs[irq];
		}
		
		u32 get_irqnum(InterruptSlot *slot)
		{
			assert(slot - irqs >= 0 && slot - irqs < (int)num_irqs);
			return slot - irqs;
		}
	};
	
	struct Interrupt {
		Interrupt *next;
		InterruptController *controller;
		InterruptSlot *slot;
		Device device;
		
		virtual ~Interrupt()
		{
		}
		
		// Returns true if there was an interrupt condition present
		// on the device; false otherwise.  This is used to detect
		// stuck interrupts.
		
		virtual bool action() = 0;

#if 0	
		#include "core/irq-server/IRQ/Interrupt.h"
		
		Interrupt()
		{
			init_iface();
		}
#endif
		
		void get_device(Device *dev)
		{
			if (dev)
				*dev = device;
		}
		
		void get_controller(IInterruptController *con)
		{
//			if (con)
//				*con = parent;
		}
		
		void get_num(u32 *irqnum)
		{
			if (irqnum)
				*irqnum = controller->get_irqnum(slot);
		}
	};
	
	// FIXME: per-CPU
	extern bool in_irq;
}

#endif
