// mem/pagetable.cc -- Generic page table implementation
//
// Most architectures should be able to use this as is, though
// architectures with weird paging hardware can provide their own implementation.
//
// OPT: Dynamically adjust the number of pagetable levels for PTEs that
// support it (mainly for generic-pte).
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/mem.h>
#include <kern/pagealloc.h>
#include <kern/generic-pagetable.h>
#include <kern/generic-pte.h>
#include <lowlevel/atomic.h>
#include <util/misc.h>

namespace Mem {
	PageTable *kernel_page_table;

	// The architecture must specify at least one data structure
	// representing one PTE.  Non-directory PTEs must support the |
	// operation, either through overloading or providing a constructor and
	// conversion operator for an integral type.  It is assumed that a PTE
	// of zero (by using bzero on the table) is a reasonable invalid PTE,
	// but a PTE-specific bulk zero method could be added if necessary.
	//
	// Eventually multiple PTE formats will need to be supported in
	// order to dynamically choose between PAE and non-PAE on 32-bit
	// x86.  When that happens, the arch will instantiate its choice
	// of PageTableImpl template rather than export Arch::PTE.

	// A PTE must support typedefs PTE::PhysAddr and PTE::VirtAddr which
	// refers to integer types of the same size as the supported physical
	// and virtual addresses, respectively...

	// A PTE must support a typedef PTE::DirPTE which is used for
	// non-final page tables.  DirPTE may be the same as PTE.  DirPTE must
	// support valid_pte, set_pte, and shift_per_level as in a normal PTE,
	// and must implement the following methods:
	//
	// void *get_table()
	//
	// A function to return the virtual address of the table pointed to
	// by this DirPTE entry.
	//
	// static DirPTE set_table(void *table)
	//
	// A function to return a DirPTE pointing to the specified virtual
	// address.

	// A normal PTE must support the following methods:
	//
	// static void flags_to_pte(Mem::PTEFlags flagsin,
	//	                         Mem::PTEFlags maskin,
	//	                         PTE &flagsout,
	//	                         PTE &maskout)
	//
	// A function to turn Mem::PTEFlags into a PTE.  It also produces
	// a mask which can be used to produce a new pte by calling
	// oldpte.set_flags(mask, newpteflags).
	//
	// PTE set_flags(PTE mask, PTE flags)
	//
	// Apply the flags to the PTE according to the provided mask.
	//
	// Mem::PTEFlags pte_to_flags()
	//
	// A function to turn a PTE into Mem::PTEFlags
	//
	// static uint addr_to_offset(VirtAddr addr, int shift)
	//	
	// A function to take a virtual address and a shift level and return
	// the offset into the page table in entries.
	//
	// PhysAddr pte_to_addr() 
	//
	// A function to take a PTE and return the physical address contained
	// therein.
	//
	// static PTE addr_to_pte(PhysAddr phys)
	//
	// A function to take a physical address and return a PTE with that
	// address and no flags set.
	//
	// bool valid_pte()
	// bool dirty_pte()
	//
	// A function to return whether the PTE is valid/dirty or not.  This
	// is a shortcut to keep from having to do a pte_to_flags repeatedly. 
	// It would have been slightly cleaner to make this a method of PTE,
	// but that would require that PTE be implemented as a complex type,
	// and I'd rather leave that up to the architecture.
	//
	// void set_pte(PTE *table, uint offset)
	//
	// A function to set a PTE in a page table.  Normally, this is just
	// a simple assignment, but some architectures may need to do something
	// unusual (such as ensure atomicity if the PTE size is greater than
	// the word size).
	//
	// PTE xchg_pte(PTE *table, uint offset)
	//
	// As set_pte, but atomically reads the old PTE while setting the
	// new PTE, and returns the old one.
	
	// A PTE must have the following constants:
	//
	// shift_per_level:
	// The number of bits of virtual address space represented by one
	// level of page tables.  This is log2(number of pages per table).
	//
	// num_levels:
	// The number of page table levels; this is used, but not imported,
	// due to a conflict with PageTable::num_levels.
	//
	// page_size: the size of a page
	// page_shift: log2(page_size)
	//
	// kmap_start, kmap_end:
	// The kernel mappings portion of the address space is mapped into all
	// address spaces, using shared page tables.  This sharing occurs at
	// the top-level page table (hopefully this will work for all
	// architectures; it can be made configurable, at the cost of some
	// complexity).  kmap_start and kmap_end are indices into the top
	// level page table that define which region is shared.  These are only
	// relevant for process address spaces.

	using Util::round_up;

	template<typename PTE>
	void PageTableImpl<PTE>::end_map(RegionWithOffset region, PTE flags,
	                                 void *table)
	{
		uint start = PTE::addr_to_offset(region.start, PTE::page_shift);
		uint end = PTE::addr_to_offset(region.end, PTE::page_shift);
	
		Page *page = kvirt_to_page(table);
		
		assert(start < pages_per_table());
		assert(end < pages_per_table());
		assert(page->flags & Page::InUse);

		PTE *ptable = static_cast<PTE *>(table);
	
		for (uint i = start; i <= end; i++) {
			PTE newpte = PTE::addr_to_pte(region.offset) | flags;
			PTE oldpte = newpte.xchg_pte(ptable, i);

			retain_if_phys(region.offset);

			if (oldpte) {
				// vaddr is only for process aspaces, so don't worry
				// about the ulong.
				
				ulong vaddr = (ulong)region.start + 
				              ((i - start) << PTE::page_shift);
			
				kill_pte(vaddr, oldpte.pte_to_addr(),
				         oldpte.dirty_pte(), oldpte.valid_pte());
			} else {
				page->retain();
			}

			region.offset += PTE::page_size;
		}
	}

	template<typename PTE>
	void PageTableImpl<PTE>::rec_map(RegionWithOffset region, PTE flags,
	                                 void *table, int shift)
	{
		if (shift < lastlevel_shift) {
			assert(shift + DirPTE::shift_per_level - PTE::shift_per_level == PTE::page_shift);
			end_map(region, flags, table);
			return;
		}
	
		Page *page = kvirt_to_page(table);
		
		DirPTE *dtable = static_cast<DirPTE *>(table);
		uint start = DirPTE::addr_to_offset(region.start, shift);
		uint end = DirPTE::addr_to_offset(region.end, shift);
		u64 orig_end = region.end;

		assert(start < pages_per_dtable());
		assert(end < pages_per_dtable());
		assert(page->flags & Page::InUse);

		if (start != end)
			region.end = round_up(region.start + 1, shift) - 1;

		for (uint i = start; i <= end; i++) {
			void *subtable;
			
			if (!dtable[i].valid_pte()) {
				subtable = Mem::alloc_pages(1);
				bzero(subtable, PTE::page_size);
				DirPTE newpte = DirPTE::set_table(subtable);
				newpte.set_pte(dtable, i);
				page->retain();
			} else {
				subtable = dtable[i].get_table();
			}
			
			rec_map(region, flags, subtable, shift - DirPTE::shift_per_level);
			
			region.offset += region.end - region.start + 1;
			region.start = region.end + 1;
			
			if (i + 1 == end)
				region.end = orig_end;
			else
				region.end += 1UL << shift;
		}
	}

	template <typename PTE>
	void PageTableImpl<PTE>::end_unmap(Region region, void *table)
	{
		Page *page = kvirt_to_page(table);
		uint start = PTE::addr_to_offset(region.start, PTE::page_shift);
		uint end = PTE::addr_to_offset(region.end, PTE::page_shift);

		assert(start < pages_per_table());
		assert(end < pages_per_table());
		assert(page->flags & Page::InUse);

		PTE *ptable = static_cast<PTE *>(table);
	
		for (uint i = start; i <= end; i++) {
			if (ptable[i]) {
				PTE oldpte = PTE().xchg_pte(ptable, i);

				if (oldpte) {
					// vaddr is only for process aspaces, so don't worry
					// about the ulong.
					
					ulong vaddr = (ulong)region.start + 
					              ((i - start) << PTE::page_shift);
				
					kill_pte(vaddr, oldpte.pte_to_addr(),
					         oldpte.dirty_pte(), oldpte.valid_pte());
				}
	
				assert(page->inuse.refcount > 1);
				page->release();
			}
		}
	}
	
	template <typename PTE>
	void PageTableImpl<PTE>::rec_unmap(Region region, void *table, int shift)
	{
		if (shift < lastlevel_shift) {
			assert(shift + DirPTE::shift_per_level - PTE::shift_per_level == PTE::page_shift);
			end_unmap(region, table);
			return;
		}

		Page *page = kvirt_to_page(table);
		uint start = DirPTE::addr_to_offset(region.start, shift);
		uint end = DirPTE::addr_to_offset(region.end, shift);
		u64 orig_end = region.end;

		assert(start < pages_per_dtable());
		assert(end < pages_per_dtable());
		assert(page->flags & Page::InUse);

		DirPTE *dtable = static_cast<DirPTE *>(table);
	
		if (start != end)
			region.end = round_up(region.start + 1, shift) - 1;

		for (uint i = start; i <= end; i++) {
			if (dtable[i].valid_pte()) {
				void *subtable = dtable[i].get_table();
			
				rec_unmap(region, subtable, shift - DirPTE::shift_per_level);
				
				Page *subpage = kvirt_to_page(subtable);
				assert(subpage->flags & Page::InUse);
				assert(subpage->inuse.refcount > 0);
				
				if (subpage->inuse.refcount == 1) {
					DirPTE().set_pte(dtable, i);
					subpage->release();

					assert(page->inuse.refcount > 1);
					page->release();
				}
			}
			
			region.start = region.end + 1;
			
			if (i + 1 == end)
				region.end = orig_end;
			else
				region.end += 1UL << shift;
		}
	}
	
	template <typename PTE>
	void PageTableImpl<PTE>::end_set_flags(Region region, PTE flags,
	                                       PTE mask, void *table)
	{
		uint start = PTE::addr_to_offset(region.start, PTE::page_shift);
		uint end = PTE::addr_to_offset(region.end, PTE::page_shift);

		assert(start < pages_per_table());
		assert(end < pages_per_table());

		PTE *ptable = static_cast<PTE *>(table);
	
		for (uint i = start; i <= end; i++) {
			if (ptable[i]) {
				PTE oldpte = ptable[i].set_flags(mask, flags);
				
				// vaddr is only for process aspaces, so don't worry
				// about the ulong.
				
				ulong vaddr = (ulong)region.start + 
				              ((i - start) << PTE::page_shift);
			
				kill_pte(vaddr, oldpte.pte_to_addr(),
				         oldpte.dirty_pte(), oldpte.valid_pte(), true);
			}
		}
	}

	template <typename PTE>
	void PageTableImpl<PTE>::rec_set_flags(Region region, PTE flags,
	                                       PTE mask, void *table,
	                                       int shift)
	{
		if (shift < lastlevel_shift) {
			assert(shift + DirPTE::shift_per_level - PTE::shift_per_level == PTE::page_shift);
			shift = PTE::page_shift;
		}

		uint start = DirPTE::addr_to_offset(region.start, shift);
		uint end = DirPTE::addr_to_offset(region.end, shift);
		u64 orig_end = region.end;

		assert(start < pages_per_dtable());
		assert(end < pages_per_dtable());

		DirPTE *dtable = static_cast<DirPTE *>(table);
	
		if (start != end)
			region.end = round_up(region.start + 1, shift) - 1;

		for (uint i = start; i <= end; i++) {
			if (dtable[i].valid_pte()) {
				void *subtable = dtable[i].get_table();
			
				rec_set_flags(region, flags, mask, subtable, 
				              shift - DirPTE::shift_per_level);
			}
			
			region.start = region.end + 1;
			
			if (i + 1 == end)
				region.end = orig_end;
			else
				region.end += 1UL << shift;
		}
	}

	template <typename PTE>
	void PageTableImpl<PTE>::map(RegionWithOffset region, Flags flags)
	{
		Lock::AutoLock autolock(lock);
		PTE pte, ptemask;
		PTE::flags_to_pte(flags, ~0UL, pte, ptemask);
		PTE *table = static_cast<PTE *>(toplevel);
		rec_map(region, pte, table, toplevel_shift);
	}

	template <typename PTE>
	void PageTableImpl<PTE>::unmap(Region region)
	{
		Lock::AutoLock autolock(lock);
		PTE *table = static_cast<PTE *>(toplevel);
		rec_unmap(region, table, toplevel_shift);
	}

	template <typename PTE>
	void PageTableImpl<PTE>::set_flags(Region region, Flags flags, Flags mask)
	{
		Lock::AutoLock autolock(lock);
		PTE pte, ptemask;
		PTE::flags_to_pte(flags, mask, pte, ptemask);
		PTE *table = static_cast<PTE *>(toplevel);
		rec_set_flags(region, pte, ptemask, table, toplevel_shift);
	}
	
	template <typename PTE>
	void PageTableImpl<PTE>::get_mapping(u64 addr, u64 *phys, Flags *flags)
	{
		Lock::AutoLock autolock(lock);
		int shift = toplevel_shift;
		
		void *table = toplevel;
		
		while (shift >= lastlevel_shift) {
			DirPTE *dtable = static_cast<DirPTE *>(table);
			DirPTE dpte = dtable[DirPTE::addr_to_offset(addr, shift)];
			
			if (!dpte.valid_pte()) {
				flags->Valid = 0;
				return;
			}
			
			table = dpte.get_table();
			shift -= DirPTE::shift_per_level;
		}
		
		assert(shift + DirPTE::shift_per_level - PTE::shift_per_level == PTE::page_shift);

		PTE *ptable = static_cast<PTE *>(table);
		
		uint off = PTE::addr_to_offset(addr, PTE::page_shift);

		*phys = ptable[off].pte_to_addr();
		*flags = ptable[off].pte_to_flags();
	}

	template <typename PTE>
	PageTableImpl<PTE>::PageTableImpl(bool process) : PageTable(process)
	{
		toplevel = Mem::alloc_pages(1);
		PTE *table = static_cast<PTE *>(toplevel);
		
		if (is_process) {
			num_levels = PTE::num_levels;
		
			if (PTE::kmap_start != 0)
				bzero(table, PTE::kmap_start * sizeof(PTE));
	
			if (PTE::kmap_end != pages_per_dtable() - 1)
				bzero(table + PTE::kmap_end + 1, 
				      (pages_per_dtable() - PTE::kmap_end - 1) * sizeof(PTE));

			PTE *ktable = static_cast<PTE *>(kernel_page_table->toplevel);
	
			memcpy(table + PTE::kmap_start, ktable + PTE::kmap_start,
			       (PTE::kmap_end - PTE::kmap_start + 1) * sizeof(PTE));
		} else {
			// FIXME: growable levels
			num_levels = PTE::num_levels;
			bzero(table, PTE::page_size);
		}

		toplevel_shift = lastlevel_shift = PTE::page_shift;
			
		if (num_levels > 1) {
			lastlevel_shift += PTE::shift_per_level;
			toplevel_shift += PTE::shift_per_level +
				               (num_levels - 2) * DirPTE::shift_per_level;
		}
	}
	
	template <typename PTE>
	PageTableImpl<PTE>::PageTableImpl(void *table) : PageTable(true)
	{
		assert(!kernel_page_table);
	
		toplevel = table;
		num_levels = PTE::num_levels;
		kernel_page_table = this;
	}
	
	template <typename PTE>
	PageTableImpl<PTE>::~PageTableImpl()
	{
		assert(this != kernel_page_table);
	
		if (is_process) {
			Region region1 = { 0, ((VirtAddr)PTE::kmap_start - 1) << toplevel_shift };
			Region region2 = { ((VirtAddr)PTE::kmap_end + 1) << toplevel_shift, ~0UL };
			
			if (PTE::kmap_start != 0)
				unmap(region1);
			if (PTE::kmap_end != pages_per_dtable() - 1)
				unmap(region2);
		} else {
			Region region = { 0, ~0UL };
			unmap(region);
		}

		Page *page = kvirt_to_page(toplevel);
		assert(page->flags & Page::InUse);
		assert(page->inuse.refcount == 1);

		Mem::free_pages(toplevel, 1);
	}
	
	template class PageTableImpl<Arch::PTE>;
	template class PageTableImpl<GenPTE>;
}
